package extensions.reports

import javax.inject.Inject
import com.typesafe.config.Config
import extensions.reports.generic.{DateTimeFormField, InputFormField, TimeFormField, TimeZoneFormField}
import net.sf.jasperreports.engine.JRParameter
import play.api.Logger
import play.api.libs.Files.TemporaryFile
import play.api.mvc.{Action, AnyContent, MultipartFormData, Request}
import utils.reports.{FieldProperties, ReverseRouting}

import scala.util.Try

class GenericJasperReportDefinition
  extends JasperReportDefinition with GenericFormSupport {

  private val logger = Logger(getClass)

  private var reportDefinitionLoader: ReportDefinitionLoader = null

  @Inject
  override var reverseRouting: ReverseRouting = null

  protected def getReportDefinitionLoader: ReportDefinitionLoader = {
    reportDefinitionLoader
  }

  protected def createField(jrParameter: JRParameter): GenericFormField = {
    Option(jrParameter.getPropertiesMap).flatMap(m => Option(m.getProperty(FieldProperties.FIELD_CLASS))) match {
      case Some(fieldClassName) =>
        Try(reportDefinitionLoader.getClassLoader.loadClass(fieldClassName)).toOption match {
          case Some(clazz) =>
            val field = clazz.newInstance().asInstanceOf[JasperFormField]
            field.init(jrParameter)
            field
          case _ =>
            logger.error(s"no class registered for name=$fieldClassName; unable to register form field")
            throw new NotImplementedError("")
        }
      case _ =>
        // old-style parameter field mapping
        genericFieldFactory.get(jrParameter.getValueClass) match {
          case Some(factory) =>
            factory(jrParameter)
          case _ => // FIXME
            throw new NotImplementedError("")
        }
    }
  }

  protected val genericFieldFactory = Map[AnyRef, PartialFunction[AnyRef, GenericFormField]](
    classOf[java.lang.String] -> {
      case p: JRParameter =>
        if (p.getName.endsWith("_TIME")) {
          new TimeFormField(p.getName, p.getDescription, {
            val defaultExpr = p.getDefaultValueExpression
            if (defaultExpr != null && defaultExpr.getText != "") {
              defaultExpr.getText.replaceAll("^\"|\"$", "")
            } else {
              "12:00"
            }
          })
        } else {
          new InputFormField(p.getName, p.getDescription, "text", {
            val defaultExpr = p.getDefaultValueExpression
            if (defaultExpr != null && defaultExpr.getText != "") {
              defaultExpr.getText.replaceAll("^\"|\"$", "")
            } else {
              ""
            }
          })
        }
    },
    classOf[java.lang.Integer] -> {
      case p: JRParameter => new InputFormField(p.getName, p.getDescription, "number", {
        val defaultExpr = p.getDefaultValueExpression
        if (defaultExpr != null && defaultExpr.getText != "") {
          defaultExpr.getText
        } else {
          ""
        }
      })
    },
    classOf[java.sql.Timestamp] -> {
      case p: JRParameter =>
        if (p.getName.endsWith("_DATETIME")) {
          new DateTimeFormField(p.getName, p.getDescription)
        } else {
          new DateTimeFormField(p.getName, p.getDescription)
        }
    },
    classOf[java.util.TimeZone] -> {
      case p: JRParameter =>
        val f = new TimeZoneFormField()
        f.init(p)
        f
    }
  )

  override def init(config: Config, loader: ReportDefinitionLoader): Unit = {
    super.init(config, loader)

    reportDefinitionLoader = loader // definition loader is passed via field to keep class interface not changed

    mainReport.getParameters
      .filter(p => (p.isForPrompting && !p.isSystemDefined) || p.getName == "REPORT_TIME_ZONE")
      .foreach(p => {
        val field = createField(p)
        registerField(field)
      })
  }

  override def key: String = {
    Try(config.getString("definition.generic.key")).getOrElse(title.hashCode.toString)
  }

  override def title: String = {
    Try(config.getString("definition.generic.title")).getOrElse(super.title)
  }

  override def renderFormAction: Action[AnyContent] = super.renderFormAction

  override def processForm(request: Request[MultipartFormData[TemporaryFile]]): AnyRef = super.processForm(request)
}
