package extensions.reports

import java.io.{File, FileNotFoundException, FileOutputStream}
import java.util
import java.util.UUID
import javax.sql.DataSource

import com.typesafe.config.Config
import net.sf.jasperreports.engine.export.JRXlsExporter
import net.sf.jasperreports.engine.util.JRLoader
import net.sf.jasperreports.engine.{JRParameter, JasperExportManager, JasperFillManager, JasperReport}
import net.sf.jasperreports.export.{SimpleXlsExporterConfiguration, SimpleExporterInput, SimpleOutputStreamExporterOutput, SimpleXlsReportConfiguration}
import org.joda.time.DateTime
import org.joda.time.format.DateTimeFormat
import play.api.Logger

import scala.collection.JavaConversions._
import scala.util.Try

abstract class JasperReportDefinition extends ReportDefinition {
  protected var config: Config = null

  private var definitionDir: String = null

  private var mainDesignFile: String = null
  private var htmlDesignFile: String = null

  private var htmlPaging: Boolean = false

  protected var mainReport: JasperReport = null
  protected var htmlReport: JasperReport = null

  private val logger = Logger(getClass)

  override def init(config: Config, loader: ReportDefinitionLoader): Unit = {
    this.config = config
    this.definitionDir = loader.getDefinitionDirectory

    htmlPaging = Try(config.getBoolean("definition.htmlPaging")).getOrElse(false)

    mainDesignFile = Try(config.getString("definition.jasper.design.main")).getOrElse("report.jasper")
    loader.getResourceAsStream(mainDesignFile) match {
      case Some(stream) =>
        logger.debug(s"loading $mainDesignFile template")
        try {
          mainReport = JRLoader.loadObject(stream).asInstanceOf[JasperReport]
        } finally {
          stream.close()
        }
        logger.info(s"loaded $mainDesignFile template")
      case _ =>
        logger.warn(s"no $mainDesignFile file found")
        throw new FileNotFoundException(s"no $mainDesignFile file found")
    }

    htmlDesignFile = Try(config.getString("definition.jasper.design.html")).getOrElse("report-html.jasper")
    loader.getResourceAsStream(htmlDesignFile) match {
      case Some(stream) =>
        logger.debug(s"loading $htmlDesignFile template")
        try {
          htmlReport = JRLoader.loadObject(stream).asInstanceOf[JasperReport]
        } finally {
          stream.close()
        }
        logger.info(s"loaded $htmlDesignFile template for HTML-version")
      case _ =>
        logger.info(s"no $htmlDesignFile file found for HTML-version; using main template instead")
        htmlReport = mainReport
    }
  }

  override def title: String = {
    mainReport.getName
  }

  override def description: String = {
    ""
  }

  override def generateReport(format: String, formObject: AnyRef, dataSource: DataSource, to: File): File = {
    val form = formObject.asInstanceOf[Map[String, AnyRef]]
    val file = generateJasperReport(format, form, dataSource, to)
    file
  }

  protected def getClassLoader: ClassLoader = {
    getClass.getClassLoader
  }

  private val fileNameDtFormatter = DateTimeFormat.forPattern("dd-MM-yyyy HH-mm");

  private def titleToFileName(title: String): String = {
    var t = title.replace("\\", "_")
    t = t.replace("/", "_")
    t = t.replace(".", "_")
    s"$t (${fileNameDtFormatter.print(DateTime.now())})__tmp"
  }

  protected def generateJasperReport(format: String, data: Map[String, AnyRef],
                                     dataSource: DataSource, to: File): File = {
    val hashMap = new util.HashMap[String, AnyRef]()
    data.entrySet().foreach(entry => {
      hashMap.put(entry.getKey, entry.getValue)
    })

    hashMap.put("DEFINITION_DIR", definitionDir)
    hashMap.put("SQL_DATA_SOURCE", dataSource)

    if (format == "html") {
      hashMap.put(JRParameter.REPORT_MAX_COUNT, new Integer(1000))
    } else {
      hashMap.put(JRParameter.REPORT_MAX_COUNT, new Integer(50000))
    }
    if (format == "xls") {
      hashMap.put(JRParameter.IS_IGNORE_PAGINATION, new java.lang.Boolean(true))
    }

    val classLoader = Thread.currentThread().getContextClassLoader
    Thread.currentThread().setContextClassLoader(getClassLoader)

    val sqlConnection = dataSource.getConnection
    try {
      val jasperPrint = JasperFillManager.fillReport(if (format == "html") { htmlReport } else { mainReport }, hashMap, sqlConnection)

      logger.debug(s"getPages.size()=${jasperPrint.getPages.size}, htmlPaging=$htmlPaging, format=$format")

      if (htmlPaging && format == "html" && jasperPrint.getPages.isEmpty) {
        null // no pages — no report
      } else {
        val file = if (format == "html") {
          new File(s"${to.getAbsolutePath}${File.separator}${UUID.randomUUID()}.$format")
        } else {
          new File(s"${to.getAbsolutePath}${File.separator}${titleToFileName(title)}.$format")
        }
        try {
          format match {
            case "html" =>
              JasperExportManager.exportReportToHtmlFile(jasperPrint, file.getAbsolutePath)
            case "pdf" =>
              JasperExportManager.exportReportToPdfFile(jasperPrint, file.getAbsolutePath)
            case "xls" =>
              val reportConfiguration = new SimpleXlsReportConfiguration()
              reportConfiguration.setOnePagePerSheet(true)
              reportConfiguration.setDetectCellType(true)
              reportConfiguration.setCollapseRowSpan(false)
              reportConfiguration.setWhitePageBackground(false)

              val exporterConfiguration = new SimpleXlsExporterConfiguration()
              exporterConfiguration.setCreateCustomPalette(true)

              val stream = new FileOutputStream(file, true)
              try {
                val exporter = new JRXlsExporter()
                exporter.setConfiguration(reportConfiguration)
                exporter.setConfiguration(exporterConfiguration)
                exporter.setExporterInput(new SimpleExporterInput(jasperPrint))
                exporter.setExporterOutput(new SimpleOutputStreamExporterOutput(stream))
                exporter.exportReport()
              } finally {
                stream.flush()
                stream.close()
              }
            case _ =>
              throw new NotImplementedError(s"format $format not supported")
          }

          file
        } catch {
          case t: Throwable =>
            file.delete()
            throw t
        }
      }
    } finally {
      sqlConnection.close()
      Thread.currentThread().setContextClassLoader(classLoader)
    }

  }

  override def getProperty(key: String): Option[String] = {
    Option(mainReport.getProperty(key))
  }
}
